# -*- coding: utf-8 -*-
'''
Created on Sun Dec  8 14:49:19 2024

@author: mstep
'''

# See: https://stackoverflow.com/questions/55619783/tkinter-i-created-a-custom-scrollable-frame-class-is-it-possible-for-it-to-au

import tkinter as tk

class ScrolledFrame(tk.Frame):
    def __init__(self, frame, *args, **kwargs):
        tk.Frame.__init__(self, frame, *args, **kwargs)

        # Make row 0 and column 0 expand.
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        # Scrollbars.
        self.v_scroll = tk.Scrollbar(self, orient=tk.VERTICAL)
        self.v_scroll.grid(row=0, column=1, sticky=tk.NSEW)

        self.h_scroll = tk.Scrollbar(self, orient=tk.HORIZONTAL)
        self.h_scroll.grid(row=1, column=0, sticky=tk.NSEW)

        # Canvas.
        self.canvas = tk.Canvas(self,
            xscrollcommand=self.h_scroll.set,
            yscrollcommand=self.v_scroll.set)
        self.canvas.grid(row=0, column=0, sticky=tk.NSEW)

        # Configure the canvas.
        self.v_scroll.config(command=self.canvas.yview)
        self.h_scroll.config(command=self.canvas.xview)

        # Frame.
        self.frame = tk.Frame(self.canvas)
        self.frame_window = self.canvas.create_window(0, 0,
            window=self.frame, anchor=tk.NW)

    def configure_frame(self):
        self.frame.update_idletasks()
        width = self.frame.winfo_reqwidth()
        height = self.frame.winfo_reqheight()
        self.canvas.config(scrollregion=(0, 0, width, height))

        # Bind the mouse wheel to scroll the canvas.
        for child in widget_and_descendants(self.canvas):
            child.bind('<MouseWheel>', self.wheel_scroll)
            #child.bind('<Button-4>', self.wheel_scroll)
            #child.bind('<Button-5>', self.wheel_scroll)

    def wheel_scroll(self, event):
        ''' Scroll the canvas.'''
        # This is the Windows version.
        # Divide event.delta by some number to slow scrolling. Some people
        # recommend dividing by 120 for Windows, but I left it alone to make
        # it scroll about the same speed when the mouse is over the scrollbar.
        units = -int(event.delta / 30)
        self.canvas.yview_scroll(units, 'units')

def widget_and_descendants(parent):
    '''Return this widget and all of its descendants.'''
    results = [parent]
    for child in parent.winfo_children():
        results += widget_and_descendants(child)
    return results

#%%
import tkinter as tk
from tkinter import messagebox
from tkinter import filedialog as fd
from PIL import Image, ImageTk, ImageDraw
import pathlib

class ScrolledWindowApp:
    def __init__(self):
        # Initially we have no images.
        self.background_image = None
        self.overlay_image = None
        self.result_image = None
        self.current_overlay_image = None
        self.photo_images = {}        # Save non-volatile PhotoImages here.
        self.scale = 1.0              # Start at 100% scale.

        # Initially we are not drawing a line.
        self.selection_in_progress = False
        self.selection_points = None

        # Make the main interface.
        self.window = tk.Tk()
        self.window.config(background='white')
        self.window.title('pil_overlay_interactive')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('500x300')

        # Make the menus.
        self.build_menus()

        self.window.grid_rowconfigure(0, weight=1)
        self.window.grid_columnconfigure(0, weight=1)
        self.window.grid_columnconfigure(2, weight=1)

        # Overlay area.
        self.overlay_frame = ScrolledFrame(self.window)
        self.overlay_frame.grid(row=0, column=0, sticky=tk.NSEW, padx=5, pady=5)
        frame = self.overlay_frame.frame
        self.overlay_canvas = tk.Canvas(frame, width=10, height=10,
            cursor='tcross')
        self.overlay_canvas.pack(anchor=tk.NW)
        self.overlay_frame.configure_frame()

        # Button.
        self.overlay_button = tk.Button(self.window, text='>', width=3,
            command=self.perform_overlay, state='disabled')
        self.overlay_button.grid(row=0, column=1, padx=5)

        # Result area.
        self.result_frame = ScrolledFrame(self.window)
        self.result_frame.grid(row=0, column=2, sticky=tk.NSEW, padx=5, pady=5)
        frame = self.result_frame.frame
        self.result_canvas = tk.Canvas(frame, width=10, height=10)
        self.result_canvas.pack(anchor=tk.NW)
        self.result_frame.configure_frame()

        # Track buttone events on the overlay canvas.
        self.overlay_canvas.bind("<Button-1>", self.mouse_down)
        self.overlay_canvas.bind("<B1-Motion>", self.mouse_move)
        self.overlay_canvas.bind("<ButtonRelease-1>", self.mouse_up)

        # Display the window.
        self.window.focus_force()
        self.window.mainloop()

    def kill_callback(self):
        self.window.destroy()

    def build_menus(self):
        '''Build the menu bar.'''
        self.menu_bar = tk.Menu(self.window)
        self.window.config(menu=self.menu_bar)
        self.build_file_menu()
        self.build_scale_menu()

    def build_file_menu(self):
        '''Build the File menu.'''
        self.file_menu = tk.Menu(self.menu_bar, tearoff=False)
        self.menu_bar.add_cascade(label='File', menu=self.file_menu)

        self.open_photo_image = ImageTk.PhotoImage(file='open.png')
        self.save_photo_image = ImageTk.PhotoImage(file='save.png')
        self.exit_photo_image = ImageTk.PhotoImage(file='exit.png')
        self.revert_photo_image = ImageTk.PhotoImage(file='undo.png')

        # Background image
        self.file_menu.add_command(label='Background Image...', accelerator='Ctrl+B',
            compound='left',
            command=self.mnu_open_background, image=self.open_photo_image)
        self.window.bind('<Control-b>', lambda event: self.mnu_open_background())

        # Overlay image
        self.file_menu.add_command(label='Overlay Image...', accelerator='Ctrl+O',
            compound='left',
            command=self.mnu_open_overlay, image=self.open_photo_image)
        self.window.bind('<Control-o>', lambda event: self.mnu_open_overlay())

        # Save
        self.file_menu.add_command(label='Save As...', accelerator='Ctrl+S',
            compound='left',
            command=self.mnu_save_as, image=self.save_photo_image,
            state='disabled')
        self.window.bind('<Control-s>', lambda event: self.mnu_save_as())

        # Revert
        self.file_menu.add_command(label='Revert', accelerator='Ctrl+R',
            compound='left',
            command=self.revert, image=self.revert_photo_image,
            state='disabled')
        self.window.bind('<Control-r>', lambda event: self.revert())

        self.file_menu.add_separator()

        # Exit
        self.file_menu.add_command(label='Exit', compound='left',
            command=self.kill_callback, image=self.exit_photo_image)

    def build_scale_menu(self):
        '''Build the Scale menu.'''
        self.scale_menu = tk.Menu(self.menu_bar, tearoff=False)
        self.menu_bar.add_cascade(label='Scale', menu=self.scale_menu)

        self.selected_scale = tk.DoubleVar(value=1)
        scales = [3, 2, 1, 0.5, 0.25, 0.1]
        for scale in scales:
            text = f'{scale:.0%}'
            self.scale_menu.add_radiobutton(label=text,
                variable=self.selected_scale, value=scale,
                command=self.set_scale)

    def mnu_open_background(self):
        image = self.get_image_file('Background Image')
        if image is None: return
        self.background_image = image
        self.revert()

    def mnu_open_overlay(self):
        image = self.get_image_file('Overlay Image')
        if image is None: return
        self.overlay_image = image
        self.current_overlay_image = self.overlay_image.copy()
        self.revert()

    def get_image_file(self, title):
        '''Let the user pick an image file. Load and return the image.'''
        filetypes = (
            ('Image Files', '*.png;*.jpg;*.bmp;*.gif;*.tif'),
            ('All files', '*.*')
        )
        filename = fd.askopenfilename(title=title, filetypes=filetypes)
        if not filename: return None
        try:
            return Image.open(filename)
        except Exception as e:
            messagebox.showinfo('Open Error', f'Error loading image file.\n{e}')
            return None

    def mnu_save_as(self):
        self.save_image_as(self.result_image)

    def save_image_as(self, image):
        '''Save a file.'''
        filetypes = (
            ('Image Files', '*.png;*.jpg;*.bmp;*.gif;*.tif'),
            ('All files', '*.*')
        )
        filename = fd.asksaveasfilename(filetypes=filetypes,
            defaultextension='.jpg')
        if len(filename) == 0:
            return

        # See what kind of file is it.
        suffix = pathlib.Path(filename).suffix.lower()
        if suffix == '.jpg' or suffix == '.jpeg':
            # It's a JPG. Save in RGB format.
            image.convert('RGB').save(filename)
        else:
            # Save in RGBA format.
            image.save(filename)

    def show_image_in_canvas(self, image, canvas, scrolled_frame, scale):
        '''Display the image in the canvas and configure the ScrolledFrame.'''
        # Delete any previus content.
        canvas.delete(tk.ALL)
        if image is None: return

        # Scale the image.
        wid = int(scale * image.width)
        hgt = int(scale * image.height)
        image = image.resize((wid, hgt))

        # Display the scaled image.
        photo_image = ImageTk.PhotoImage(image)
        self.photo_images[canvas] = photo_image
        canvas.create_image(0, 0, image=photo_image, anchor=tk.NW)
        canvas.config(width=image.width, height=image.height)

        # Configure the ScrolledFrame.
        scrolled_frame.configure_frame()

    def set_scale(self):
        self.scale = self.selected_scale.get()
        self.display_images()

    def revert(self):
        '''Remove any edits and redisplay the images.'''
        if self.background_image is None:
            self.result_image = None
        else:
            self.result_image = self.background_image.copy()
        self.draw_selection()

    def display_images(self):
        '''Display any images that we have at the current scale.'''
        self.show_image_in_canvas(
            self.current_overlay_image,
            self.overlay_canvas,
            self.overlay_frame,
            self.scale)
        self.show_image_in_canvas(
            self.result_image,
            self.result_canvas,
            self.result_frame,
            self.scale)

        # If both images are loaded, enable the Save As and Revert commands.
        if self.overlay_image is not None and \
           self.background_image is not None:
               self.file_menu.entryconfig('Save As...', state='normal')
               self.file_menu.entryconfig('Revert', state='normal')

    def mouse_down(self, event):
        '''Save the drawing start point.'''
        if self.result_image is None: return  # Bail if no images loaded.

        scale = self.selected_scale.get()
        x = int(event.x / scale)
        y = int(event.y / scale)
        self.selection_points = [(x, y)]
        self.selection_in_progress = True

    def mouse_move(self, event):
        '''Add a drawing point.'''
        if not self.selection_in_progress: return  # Bail if not selecting.

        scale = self.selected_scale.get()
        x = int(event.x / scale)
        y = int(event.y / scale)
        self.selection_points.append((x, y))

        # Draw the selection curve on the result image and display it.
        self.draw_selection()

    def mouse_up(self, event):
        '''Finish drawing.'''
        if not self.selection_in_progress: return  # Bail if not selecting.

        # We are no longer selecting.
        self.selection_in_progress = True

        # If we have at least two selection points, enable the button.
        if len(self.selection_points) >= 2:
            self.overlay_button.config(state='normal')
        else:
            # We no longer have a selection.
            self.overlay_button.config(state='disabled')
            self.selection_points = None
            self.draw_selection()

    def draw_selection(self):
        '''Draw the selection line on the result image and display it.'''
        # Start with a fresh copy of the overlay image.
        if self.overlay_image is None:
            self.current_overlay_image = None
        else:
            self.current_overlay_image = self.overlay_image.copy()

        # See if we have selection points.
        if self.selection_points is not None:
            # Draw the selection on the current overlay image.
            dr = ImageDraw.Draw(self.current_overlay_image)
            dr.line(self.selection_points, fill='yellow', width=2)

        # Display the result.
        self.display_images()

    def perform_overlay(self):
        '''Perform the overlay.'''
        # Copy the selected part of the overlay image onto the result.
        self.overlay_selected_area()

        # We no longer have a selection.
        self.selection_points = None
        self.overlay_button.config(state='disabled')
        self.draw_selection()

    def overlay_selected_area(self):
        '''Overlay the selected area onto the result image.'''
        # Make a mask image.
        wid = self.overlay_image.width
        hgt = self.overlay_image.height
        mask_image = Image.new('1', (wid, hgt), 'black')

        # White out the selected area.
        dr = ImageDraw.Draw(mask_image)
        dr.polygon(self.selection_points, fill='white', outline='white')

        # Paste the selected area from the overlay onto the background image.
        self.result_image.paste(self.overlay_image, mask=mask_image)

#%%

ScrolledWindowApp()
print('Done')
